﻿using System;
using System.Collections.Generic;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Media;
using System.Windows.Media.Media3D;
using _3DTools;

namespace Panel3D
{
  /// <summary>
  /// Interaction logic for Panel3DUserControl.xaml
  /// </summary>
  public partial class Panel3DUserControl : UserControl
  {
    readonly Brush brushdefault = new SolidColorBrush(Color.FromRgb(0xE2, 0xE2, 0xE2));
    readonly Brush brushselection = new SolidColorBrush(Colors.Gold);
    const double defaultCamLocation_x = 10;
    const double defaultCamLocation_y = 10;
    const double defaultCamLocation_z = 10;
    enum SliderFunctionality
    {
      Rotator,
      CameraLocation
    }
    #region For transforms
    private TranslateTransform3D SlidersRotator_center;
    private double CamLocation_x = defaultCamLocation_x;
    private double CamLocation_y = defaultCamLocation_y;
    private double CamLocation_z = defaultCamLocation_z;
    private AxisAngleRotation3D SlidersRotator_rot_x;
    private AxisAngleRotation3D SlidersRotator_rot_y;
    private AxisAngleRotation3D SlidersRotator_rot_z;
    private ScaleTransform3D SlidersRotator_zoom;
    private double txt_zomm_as_double
    {
      get
      {
        return GetDoubleValueOrDefault(txt_zoom.Text, 0);
      }
      set
      {
        txt_zoom.Text = value.ToString();
      }
    }
    private Transform3DGroup SlidersRotator_transform;
    private SliderFunctionality m_sliderfunctionality;
    private double DistanceFromCenter
    {
      get
      {
        return Math.Sqrt(CamLocation_x * CamLocation_x + CamLocation_y * CamLocation_y + CamLocation_z * CamLocation_z);
      }
    }
    #endregion For transforms
    private ModelVisual3DFilter filter = ModelVisual3DFilter.AllOnExceptTexts;
    private List<IModelVisual3D> elements;
    private void ChangeSliderFunctionality(SliderFunctionality functionality)
    {
      m_sliderfunctionality = functionality;
      switch (m_sliderfunctionality)
      {
        case SliderFunctionality.Rotator:
          btn_Rotator.Background = brushselection;
          btn_CameraLocation.Background = brushdefault;
          slider_x.Value = SlidersRotator_rot_x.Angle;
          slider_y.Value = SlidersRotator_rot_y.Angle;
          slider_z.Value = SlidersRotator_rot_z.Angle;
          slider_zoom.Value = SlidersRotator_zoom.ScaleX;
          break;
        case SliderFunctionality.CameraLocation:
          btn_Rotator.Background = brushdefault;
          btn_CameraLocation.Background = brushselection;
          slider_x.Value = CamLocation_x;
          slider_y.Value = CamLocation_y;
          slider_z.Value = CamLocation_z;
          slider_zoom.Value = DistanceFromCenter;
          break;
      }
    }
    public Panel3DUserControl()
    {
      SlidersRotator_Initialisation();
      InitializeComponent();
      SetCameraTransform();

      elements = new List<IModelVisual3D>();

      ClearViewport();
      SetCamera();
      AddModels();
    }
    static double GetDoubleValueOrDefault(string ValueAsString, double DefaultValue)
    {
      double ret;
      if (double.TryParse(ValueAsString, out ret))
        return ret;
      else
        return DefaultValue;
    }
    #region SlidersRotator
    private void txt_camloc_zoom_TextChanged(object sender, TextChangedEventArgs e)
    {

    }
    private void txt_camloc_x_TextChanged(object sender, TextChangedEventArgs e)
    {
      CamLocation_x = GetDoubleValueOrDefault(txt_camloc_x.Text, defaultCamLocation_x);
      SetCamera();
      if (m_sliderfunctionality == SliderFunctionality.CameraLocation)
        slider_x.Value = CamLocation_x;
    }
    private void txt_camloc_y_TextChanged(object sender, TextChangedEventArgs e)
    {
      CamLocation_y = GetDoubleValueOrDefault(txt_camloc_y.Text, defaultCamLocation_y);
      SetCamera();
      if (m_sliderfunctionality == SliderFunctionality.CameraLocation)
        slider_y.Value = CamLocation_y;
    }
    private void txt_camloc_z_TextChanged(object sender, TextChangedEventArgs e)
    {
      CamLocation_z = GetDoubleValueOrDefault(txt_camloc_z.Text, defaultCamLocation_z);
      SetCamera();
      if (m_sliderfunctionality == SliderFunctionality.CameraLocation)
        slider_z.Value = CamLocation_z;
    }
    private void slider_x_changed(object sender, RoutedEventArgs args)
    {
      TurnOffLabels();
      double val = System.Math.Round(slider_x.Value, 2);
      switch (m_sliderfunctionality)
      {
        case SliderFunctionality.Rotator:
          SlidersRotator_rot_x.Angle = val;
          if (GetDoubleValueOrDefault(txt_rot_x.Text, 0) != SlidersRotator_rot_x.Angle)
            txt_rot_x.Text = SlidersRotator_rot_x.Angle.ToString();
          break;
        case SliderFunctionality.CameraLocation:
          CamLocation_x = val;
          if (GetDoubleValueOrDefault(txt_camloc_x.Text, defaultCamLocation_x) != CamLocation_x)
            txt_camloc_x.Text = CamLocation_x.ToString();
          SetCamera();
          break;
      }
    }
    private void slider_y_changed(object sender, RoutedEventArgs args)
    {
      TurnOffLabels();
      double val = System.Math.Round(slider_y.Value, 2);
      switch (m_sliderfunctionality)
      {
        case SliderFunctionality.Rotator:
          SlidersRotator_rot_y.Angle = val;
          if (GetDoubleValueOrDefault(txt_rot_y.Text, 0) != SlidersRotator_rot_y.Angle)
            txt_rot_y.Text = SlidersRotator_rot_y.Angle.ToString();
          break;
        case SliderFunctionality.CameraLocation:
          CamLocation_y = val;
          if (GetDoubleValueOrDefault(txt_camloc_y.Text, defaultCamLocation_y) != CamLocation_y)
            txt_camloc_y.Text = CamLocation_y.ToString();
          SetCamera();
          break;
      }
    }
    private void slider_z_changed(object sender, RoutedEventArgs args)
    {
      TurnOffLabels();
      double val = System.Math.Round(slider_z.Value, 2);
      switch (m_sliderfunctionality)
      {
        case SliderFunctionality.Rotator:
          SlidersRotator_rot_z.Angle = val;
          if (GetDoubleValueOrDefault(txt_rot_z.Text, 0) != SlidersRotator_rot_z.Angle)
            txt_rot_z.Text = SlidersRotator_rot_z.Angle.ToString();
          break;
        case SliderFunctionality.CameraLocation:
          CamLocation_z = val;
          if (GetDoubleValueOrDefault(txt_camloc_z.Text, defaultCamLocation_z) != CamLocation_z)
            txt_camloc_z.Text = CamLocation_z.ToString();
          SetCamera();
          break;
      }
    }
    private void slider_zoom_changed(object sender, RoutedEventArgs args)
    {
      TurnOffLabels();
      double slider_zoom_Value = System.Math.Round(slider_zoom.Value, 2);
      switch (m_sliderfunctionality)
      {
        case SliderFunctionality.Rotator:
          SliderRotator_zoom_SetScale(slider_zoom_Value);
          if (GetDoubleValueOrDefault(txt_zoom.Text, 1) != SlidersRotator_zoom.ScaleX)
            txt_zoom.Text = SlidersRotator_zoom.ScaleX.ToString();
          break;
        case SliderFunctionality.CameraLocation:
          break;
      }
    }
    private void txt_rotx_TextChanged(object sender, TextChangedEventArgs e)
    {
      SlidersRotator_rot_x.Angle = GetDoubleValueOrDefault(txt_rot_x.Text, 0);
      if (m_sliderfunctionality == SliderFunctionality.Rotator && slider_x != null && slider_x.Value != SlidersRotator_rot_x.Angle)
        slider_x.Value = SlidersRotator_rot_x.Angle;
    }
    private void txt_rotz_TextChanged(object sender, TextChangedEventArgs e)
    {
      SlidersRotator_rot_z.Angle = GetDoubleValueOrDefault(txt_rot_z.Text, 0);
      if (m_sliderfunctionality == SliderFunctionality.Rotator && slider_z != null && slider_z.Value != SlidersRotator_rot_z.Angle)
        slider_z.Value = SlidersRotator_rot_z.Angle;
    }
    private void txt_roty_TextChanged(object sender, TextChangedEventArgs e)
    {
      SlidersRotator_rot_y.Angle = GetDoubleValueOrDefault(txt_rot_y.Text, 0);
      if (m_sliderfunctionality == SliderFunctionality.Rotator && slider_y != null && slider_y.Value != SlidersRotator_rot_y.Angle)
        slider_y.Value = SlidersRotator_rot_y.Angle;
    }
    private void SliderRotator_zoom_SetScale(double scale)
    {
      SlidersRotator_zoom.ScaleX = scale;
      SlidersRotator_zoom.ScaleY = scale;
      SlidersRotator_zoom.ScaleZ = scale;
    }
    private void txt_zoom_TextChanged(object sender, TextChangedEventArgs e)
    {
      double scale = txt_zomm_as_double;
      SliderRotator_zoom_SetScale(scale);
      if (m_sliderfunctionality == SliderFunctionality.Rotator &&
        slider_zoom != null && slider_zoom.Value != scale)
        SliderRotator_zoom_SetScale(scale);
    }
    private void SlidersRotator_Initialisation()
    {
      //SlidersRotator initialisation
      SlidersRotator_center = new TranslateTransform3D(0, 0, 0); //center point
      SlidersRotator_rot_x = new AxisAngleRotation3D(
          new Vector3D(1, 0, 0),
          0); // setup rotation angle
      SlidersRotator_rot_y = new AxisAngleRotation3D(
          new Vector3D(0, 1, 0),
          0); // setup rotation angle
      SlidersRotator_rot_z = new AxisAngleRotation3D(
          new Vector3D(0, 0, 1),
          0); // setup rotation angle
      SlidersRotator_zoom = new ScaleTransform3D(
          1,
          1,
          1);

      SlidersRotator_transform = new Transform3DGroup();

      SlidersRotator_transform.Children.Add(SlidersRotator_zoom);

      // the order of the following three is significant
      SlidersRotator_transform.Children.Add(new RotateTransform3D(SlidersRotator_rot_y));
      SlidersRotator_transform.Children.Add(new RotateTransform3D(SlidersRotator_rot_x));
      SlidersRotator_transform.Children.Add(new RotateTransform3D(SlidersRotator_rot_z));

      SlidersRotator_transform.Children.Add(SlidersRotator_center);
    }
    #endregion SlidersRotator
    private void SetCameraTransform()
    {
      mainViewport.Camera.Transform = SlidersRotator_transform;
    }
    private void ClearViewport()
    {
      if (mainViewport != null)
      {
        ModelVisual3D m;
        for (int i = mainViewport.Children.Count - 1; i >= 0; i--)
        {
          m = (ModelVisual3D)mainViewport.Children[i];
          if (m.Content is DirectionalLight == false)
            mainViewport.Children.Remove(m);
        }
      }
      if (mainViewportCanvas != null)
      {
        UIElement uielement;
        for (int i = mainViewportCanvas.Children.Count - 1; i >= 0; i--)
        {
          uielement = mainViewportCanvas.Children[i];
          mainViewportCanvas.Children.Remove(uielement);
        }
      }
    }
    private void SetCamera()
    {
      if (mainViewport != null)
      {
        PerspectiveCamera camera = (PerspectiveCamera)mainViewport.Camera;
        Point3D position = new Point3D(CamLocation_x, CamLocation_y, CamLocation_z);
        Vector3D lookDirection = new Vector3D(-CamLocation_x, -CamLocation_y, -CamLocation_z);
        camera.Position = position;
        camera.LookDirection = lookDirection;
      }
    }
    private void AddModels()
    {
      if (elements != null && mainViewport!=null)
      {
        PerspectiveCamera camera = (PerspectiveCamera)mainViewport.Camera;
        Vector3D newlookdirection = Panel3DMath.Transform3DVector(camera.Transform,camera.LookDirection);
        foreach (IModelVisual3D element in elements)
        {
          element.UpdateViewToLookDirection(newlookdirection);
          this.mainViewport.Children.Add(element.GetModelVisual3D(filter));
          this.mainViewportCanvas.Children.Add(element.GetUIElement(filter,mainViewport));
        }
      }
    }
    public List<IModelVisual3D> Elements
    {
      get
      {
        return elements;
      }
    }
    public void RefreshScene()
    {
      ClearViewport();
      SetCamera();
      AddModels();
    }
    private void btn_Rotator_Click(object sender, RoutedEventArgs e)
    {
      ChangeSliderFunctionality(SliderFunctionality.Rotator);
    }
    private void btn_CameraLocation_Click(object sender, RoutedEventArgs e)
    {
      ChangeSliderFunctionality(SliderFunctionality.CameraLocation);
    }
    #region filter checkboxes
    private void TurnOffLabels()
    {
      CheckBoxTexts2D.IsChecked = false;
      CheckBoxTexts3D.IsChecked = false;
    }
    private void CheckBoxTriangles_Checked(object sender, RoutedEventArgs e)
    {
      if (CheckBoxTriangles.IsChecked.HasValue)
      {
        if ((bool)CheckBoxTriangles.IsChecked)
          filter.Triangles = true;
        else
          filter.Triangles = false;
        RefreshScene();
      }
    }
    private void CheckBoxLines_Checked(object sender, RoutedEventArgs e)
    {
      if (CheckBoxLines.IsChecked.HasValue)
      {
        if ((bool)CheckBoxLines.IsChecked)
          filter.Lines = true;
        else
          filter.Lines = false;
        RefreshScene();
      }
    }
    private void CheckBoxLables3D_Checked(object sender, RoutedEventArgs e)
    {
      if (CheckBoxTexts3D.IsChecked.HasValue)
      {
        if ((bool)CheckBoxTexts3D.IsChecked)
          filter.Texts3D = true;
        else
          filter.Texts3D = false;
        RefreshScene();
      }
    }
    private void CheckBoxLables2D_Checked(object sender, RoutedEventArgs e)
    {
      if (CheckBoxTexts2D.IsChecked.HasValue)
      {
        if ((bool)CheckBoxTexts2D.IsChecked)
          filter.Texts2D = true;
        else
          filter.Texts2D = false;
        RefreshScene();
      }
    }
    #endregion filter checkboxes
  }
}
